/* -*- C++ -*-
 * mu_state.C
 * @(#) Auxiliary routines related to states in the verifier
 *
 * Copyright (C) 1992 - 1999 by the Board of Trustees of              
 * Leland Stanford Junior University.
 *
 * License to use, copy, modify, sell and/or distribute this software
 * and its documentation any purpose is hereby granted without royalty,
 * subject to the following terms and conditions:
 *
 * 1.  The above copyright notice and this permission notice must
 * appear in all copies of the software and related documentation.
 *
 * 2.  The name of Stanford University may not be used in advertising or
 * publicity pertaining to distribution of the software without the
 * specific, prior written permission of Stanford.
 *
 * 3.  This software may not be called "Murphi" if it has been modified
 * in any way, without the specific prior written permission of David L.
 * Dill.
 *
 * 4.  THE SOFTWARE IS PROVIDED "AS-IS" AND STANFORD MAKES NO
 * REPRESENTATIONS OR WARRANTIES, EXPRESS OR IMPLIED, BY WAY OF EXAMPLE,
 * BUT NOT LIMITATION.  STANFORD MAKES NO REPRESENTATIONS OR WARRANTIES
 * OF MERCHANTABILITY OR FITNESS FOR ANY PARTICULAR PURPOSE OR THAT THE
 * USE OF THE SOFTWARE WILL NOT INFRINGE ANY PATENTS, COPYRIGHTS
 * TRADEMARKS OR OTHER RIGHTS. STANFORD SHALL NOT BE LIABLE FOR ANY
 * LIABILITY OR DAMAGES WITH RESPECT TO ANY CLAIM BY LICENSEE OR ANY
 * THIRD PARTY ON ACCOUNT OF, OR ARISING FROM THE LICENSE, OR ANY
 * SUBLICENSE OR USE OF THE SOFTWARE OR ANY SERVICE OR SUPPORT.
 *
 * LICENSEE shall indemnify, hold harmless and defend STANFORD and its
 * trustees, officers, employees, students and agents against any and all
 * claims arising out of the exercise of any rights under this Agreement,
 * including, without limiting the generality of the foregoing, against
 * any damages, losses or liabilities whatsoever with respect to death or
 * injury to person or damage to property arising from or out of the
 * possession, use, or operation of Software or Licensed Program(s) by
 * LICENSEE or its customers.
 *
 * Read the file "license" distributed with these sources, or call
 * Murphi with the -l switch for additional information.
 * 
 */

/* 
 * Original Author: Ralph Melton
 * Extracted from mu_epilog.inc and mu_prolog.inc
 * by C. Norris Ip
 * Created: 21 April 93
 *
 */ 
 
 /* 
 * Cached Murphi Release 4.1
 * Author: G. Della Penna
 * Released: 30 Jan 2003
 * Notes: Several methods were modified or added to support
 *        the disk management routines. See README.cmurphi for more
 *        information.
 *       
 */


/****************************************
  There are 3 groups of implementations:
  1) bit vector
  2) class StatePtr and state related stuff
  3) state queue and stack
  4) state set
  ****************************************/

#include <limits.h>

void
state::print()
{
  theworld.print();
};

/****************************************
  Bit vector - copied straight from Andreas. 
  ****************************************/
dynBitVec::dynBitVec( unsigned long nBits )
: numBits( nBits )
{
  v = new unsigned char[ NumBytes() ]; /* Allocate and clear vector. */
  ErrAlloc(v);
  memset( v, 0, NumBytes() );
};

dynBitVec::~dynBitVec()
{
  delete[ OLD_GPP(NumBytes()) ] v; // should be delete[].
}


/****************************************
  class StatePtr and state related stuff.
  ****************************************/

inline void StatePtr::sCheck() {
#ifdef HASHC
  if (args->trace_file.value)
    Error.Notrace("Internal: Illegal Access to StatePtr.");
#endif
}

inline void StatePtr::lCheck() {
#ifdef HASHC
  if (!args->trace_file.value)
    Error.Notrace("Internal: Illegal Access to StatePtr.");
#endif
}

StatePtr::StatePtr(state* s) { sCheck(); sp=s; }
StatePtr::StatePtr(unsigned long l) { lCheck(); lv=l; }

void StatePtr::set(state* s) { sCheck(); sp=s; }
void StatePtr::set(unsigned long l) { lCheck(); lv=l; }
void StatePtr::clear() {
#ifdef HASHC
  if (args->trace_file.value) 
    lv=0;
  else
#endif 
    sp=NULL;
} 
state* StatePtr::sVal() { sCheck(); return sp; }
unsigned long StatePtr::lVal() { lCheck(); return lv; }

StatePtr StatePtr::previous() {   // return StatePtr to previous state
#ifdef HASHC
  if (args->trace_file.value)
    return TraceFile->read(lv)->previous;
  else
#endif
    return sp->previous.sp;
}

bool StatePtr::isStart() {   // check if I point to a startstate
#ifdef HASHC
  if (args->trace_file.value) {
    if (TraceFile->read(lv)->previous==0) return TRUE;
    return FALSE;
  }
  else 
#endif
  {
    if (sp->previous.sp==NULL) return TRUE;
    return FALSE;
  }
}

bool StatePtr::compare(state* s) {   // compare the state I point to with s
#ifdef HASHC
  if (args->trace_file.value) {
    unsigned long *key = h3->hash(s, FALSE);
    unsigned long c1 = key[1] &
      ((~0UL) << (args->num_bits.value>32 ? 0 : 32-args->num_bits.value));
    unsigned long c2 = key[2] &
      (args->num_bits.value>32 ? (~0UL)<<(64-args->num_bits.value) : 0UL);

    return ( c1==TraceFile->read(lv)->c1 &&
             c2==TraceFile->read(lv)->c2 );
  }
  else
#endif
    return (StateCmp(sp,s)==0);
}


void 
StateCopy(state * l, state * r)
// Uli: uses default assignment operator
{
  *l = *r;
}

int 
StateCmp(state * l, state * r)
{
  int i = BLOCKS_IN_WORLD/4;
  register int *d = (int *)l->bits, *s = (int *)r->bits;

  while (i--)
      if( *d > *s)
	  return 1;
      else if( *d++ < *s++)
	  return -1;
  return 0;
}

void 
copy_state(state *& s)
{
  state *h;

  if ( (h = new state) == NULL )
    Error.Notrace
      ("New failed. Swap space probably too small for state queue.");
  *h = *s;
  s = h;
}

bool 
StateEquivalent(state * l, StatePtr r)
{
  return match(l, r);
}

/****************************************
  Modification:
  1) 1 Dec 93 Norris Ip: 
  check -sym option when checking was_present()
  add StateCmp(state l, state r)
  2) 24 Feb 94 Norris Ip:
  added -debugsym option to run two hash tables in parallel
  for debugging purpose
  3) 8 March 94 Norris Ip:
  merge with the latest rel2.6
****************************************/

/********************
  $Log: mu_state.C,v $
  
  Revision 2.0  2003/01/10 17:06:12  giuseppe 
  disk extension - final release

  Revision 1.7  2001/10/28 21:50:00  giuseppe 
  disk algorithm support complete - start testing

  Revision 1.6  2001/10/01 18:41:14  giuseppe 
  start adding support for disk algorithm
  
  Revision 1.5  2001/06/11 16:39:11  giuseppe 
  cache extension - final release  
  
  Revision 1.4  2001/02/28 20:32:12  enrico
  modified simple_was_present

  Revision 1.3  2001/02/05 08:34:43  enrico
  adj

  Revision 1.2  2001/02/02 09:30:57  enrico
  updated debug printing

  Revision 1.1  2001/02/02 08:44:50  enrico
  Initial revision

*******************

  Revision 1.3  1999/01/29 08:28:09  uli
  efficiency improvements for security protocols

  Revision 1.2  1999/01/29 07:49:11  uli
  bugfixes

  Revision 1.4  1996/08/07 18:54:33  ip
  last bug fix on NextRule/SetNextEnabledRule has a bug; fixed this turn

  Revision 1.3  1996/08/07 01:00:18  ip
  Fixed bug on what_rule setting during guard evaluation; otherwise, bad diagnoistic message on undefine error on guard

  Revision 1.2  1996/08/07 00:15:26  ip
  fixed while code generation bug

  Revision 1.1  1996/08/07 00:14:46  ip
  Initial revision

********************/
